"""
JSORC APIs
"""

import yaml
import json
from json import dumps, loads
from time import time
from base64 import b64decode
from jaseci.svc import CommonService
from jaseci.svc.common import Kube, UNSAFE_PARAPHRASE
from jaseci.api.interface import Interface


class JsOrcApi:
    """
    API for managing JsOrc
    """

    @Interface.admin_api()
    def load_yaml(self, files: list, namespace: str = "default"):
        """
        applying list of yaml files without associating to any modules/services
        """

        if self._h.meta.is_automated():
            kube = self._h.meta.app.kubernetes
            kube: Kube

            res = {}

            for file in files:
                for conf in yaml.safe_load_all(b64decode(file["base64"])):
                    kind = conf["kind"]
                    kube.create(kind, namespace, conf)
                    if not res.get(kind):
                        res[kind] = []
                    res[kind].append(conf)

            return res
        else:
            return {"message": "load_yaml is not supported on non automated JsOrc!"}

    @Interface.admin_api(cli_args=["name"])
    def apply_yaml(self, name: str, file: list, unsafe_paraphrase: str = ""):
        """
        apply manifest yaml to specific service
        """

        new_config = {}

        config_version = str(time())

        for conf in yaml.safe_load_all(b64decode(file[0]["base64"])):
            kind = conf["kind"]
            labels = conf.get("metadata").get("labels")
            if not labels.get("config_version"):
                labels["config_version"] = config_version

            if not new_config.get(kind):
                new_config[kind] = []
            new_config[kind].append(conf)

        old_config = self._h.get_glob(name)
        if old_config:
            old_config = loads(old_config)
            old_config.pop("__OLD_CONFIG__", None)
            for kind, confs in old_config.items():
                names = []
                for conf in confs:
                    names.append(conf["metadata"]["name"])
                old_config[kind] = names

            new_config["__OLD_CONFIG__"] = old_config

            if unsafe_paraphrase == UNSAFE_PARAPHRASE:
                new_config["__UNSAFE_PARAPHRASE__"] = unsafe_paraphrase

        self._h.save_glob(name, dumps(new_config))

        return new_config

    @Interface.admin_api(cli_args=["name"])
    def service_refresh(self, name: str):
        """
        refreshing service's config. If JsOrc is not automated, service will restart else JsOrc will handle the rest
        """

        hook = self._h

        to_start = not hook.meta.is_automated()

        service = getattr(hook, name, None)

        response = {"success": False}

        if isinstance(service, CommonService):
            service.reset(hook, to_start)
            response["success"] = True
        else:
            response[
                "message"
            ] = f"{name} is not a valid service. Can not refresh config."

        return response

    @Interface.admin_api(cli_args=["name"])
    def service_call(self, svc: str, attrs: list = []):
        """
        temporary api for retreiving/calling attributes of specific instance.
        """

        from jaseci.svc import MetaService

        meta = self._h.meta
        meta: MetaService

        svc = meta.get_service(svc)

        if not svc:
            return "Service (svc) field is required!"

        if not attrs:
            return "Attributes (attrs) field is required!"

        res = svc
        for at in attrs:
            attr = at.get("attr", False)
            if attr:
                res = getattr(res, attr)

            if callable:
                args = at.get("args", [])
                kwargs = at.get("kwargs", {})
                res = res(*args, **kwargs)

        try:
            # test if json serializable
            json.dumps(res)
            return res
        except Exception:
            return {"error": f"Not JSON Serializable! class {res.__class__.__name__}"}

    @Interface.admin_api(cli_args=["name"])
    def jsorc_actions_load(self, name: str, mode: str):
        """
        Load an action as remote or local mode through JSORC.
        JSORC will load the corresponding module or start a microservice if needed.
        Return the current status of the action.
        """
        hook = self._h
        if hook.meta.run_svcs:
            hook.meta.app.load_actions(name, mode)
            status = hook.meta.app.get_actions_status(name)
            return {"success": True, "action_status": status}
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api(cli_args=["name"])
    def jsorc_actions_status(self, name: str):
        """
        Get the current status of an action
        """
        hook = self._h
        if hook.meta.run_svcs:
            status = hook.meta.app.get_actions_status(name)
            return {"success": True, "action_status": status}
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api(cli_args=["name"])
    def jsorc_actions_unload(
        self, name: str, mode: str = "auto", retire_svc: bool = True
    ):
        """
        Unload an action through JSORC.
        If retire_svc is set to True (true by default), it will also retire the corresponding microservice.
        """
        hook = self._h
        if hook.meta.run_svcs:
            res = hook.meta.app.unload_actions(name, mode, retire_svc)
            return {"success": res[0], "message": res[1]}
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api(cli_args=["config", "name"])
    def jsorc_actions_config(self, config: str, name: str):
        """
        Loading the config of an action module
        """
        hook = self._h
        if hook.meta.run_svcs:
            res = hook.meta.app.load_action_config(name, config)
            return {"success": res}
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api()
    def jsorc_trackact_start(self):
        """ "
        Instruct JSORC to start tracking any changes in actions state
        """
        hook = self._h
        if hook.meta.run_svcs:
            hook.meta.app.actions_tracking_start()
            return {"success": True}
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api()
    def jsorc_trackact_stop(self):
        """ "
        Instruct JSORC to start tracking any changes in actions state
        """
        hook = self._h
        if hook.meta.run_svcs:
            return hook.meta.app.actions_tracking_stop()
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api()
    def jsorc_benchmark_start(self):
        """
        Tell JSORC to start collecting request performance metrics
        """
        hook = self._h
        if hook.meta.run_svcs:
            hook.meta.app.benchmark_start()
            return {"success": True}
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api()
    def jsorc_benchmark_report(self):
        """
        Report the collected request performance metrics of the currently ongoing benchmark
        """
        hook = self._h
        if hook.meta.run_svcs:
            return hook.meta.app.benchmark_report()
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api()
    def jsorc_benchmark_stop(self, report: bool = True):
        """
        End the benchmark period and report performance metrics
        """
        hook = self._h
        if hook.meta.run_svcs:
            return hook.meta.app.benchmark_stop(report)
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api()
    def jsorc_tracksys_start(self):
        """
        Ask JSORC to start tracking the state of the system as observed by JSORC on every interval.
        """
        hook = self._h
        if hook.meta.run_svcs:
            hook.meta.app.state_tracking_start()
            return {"success": True}
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api()
    def jsorc_tracksys_report(self):
        """
        Report the tracked system states so far
        """
        hook = self._h
        if hook.meta.run_svcs:
            return hook.meta.app.state_tracking_report()
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api()
    def jsorc_tracksys_stop(self):
        """
        Stop state tracking for JSORC
        """
        hook = self._h
        if hook.meta.run_svcs:
            return hook.meta.app.state_tracking_stop()
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api()
    def jsorc_actionpolicy_set(self, policy_name: str, policy_params: dict = {}):
        """
        Set an action optimization policy for JSORC
        """
        hook = self._h
        if hook.meta.run_svcs:
            res = hook.meta.app.set_action_policy(policy_name, policy_params)
            if res is True:
                return {
                    "success": True,
                    "message": f"Action optimization policy configured as {policy_name} with params {policy_params}",
                }
            else:
                return {"success": False, "message": res}
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api()
    def jsorc_actionpolicy_get(self):
        """
        Get the current active action optimization policy
        """
        hook = self._h
        if hook.meta.run_svcs:
            policy = hook.meta.app.get_action_policy()
            return {"success": True, "policy": policy}
        else:
            return {"success": False, "message": "No running JSORC service."}

    @Interface.admin_api()
    def jsorc_loadtest(self, test: str, experiment: str = "", mem: int = 0):
        """
        load test API. overwritten in jaseci_serv
        """
        pass



from copy import deepcopy
from multiprocessing import Process

from kubernetes import config as kubernetes_config
from kubernetes.client import ApiClient, CoreV1Api, AppsV1Api, RbacAuthorizationV1Api
from kubernetes.client.rest import ApiException

from jaseci.utils.utils import logger
from jaseci.actions.live_actions import load_action_config
from jaseci.svc.actions_optimizer.actions_optimizer import ActionsOptimizer
from .state import ServiceState as Ss
from .config import META_CONFIG, KUBERNETES_CONFIG

import time
import numpy as np

###################################################
#                  UNSAFE PARAMS                  #
###################################################

UNSAFE_PARAPHRASE = "I know what I'm doing!"
UNSAFE_KINDS = ["PersistentVolumeClaim"]

COMMON_ERROR = "Not properly configured!"
DEFAULT_CONFIG = {"enabled": False}


class CommonService:

    _daemon = {}

    def __init__(self, hook=None):
        self.app = None
        self.enabled = False
        self.state = Ss.NOT_STARTED
        self.quiet = True
        self.build_settings(hook)

    ###################################################
    #                   PROPERTIES                    #
    ###################################################

    # ------------------- DAEMON -------------------- #

    @property
    def daemon(self):
        return __class__._daemon

    ###################################################
    #                     BUILDER                     #
    ###################################################

    def start(self, hook=None):
        try:
            if self.enabled and self.is_ready():
                self.state = Ss.STARTED
                self.run(hook)
                self.state = Ss.RUNNING
                self.post_run(hook)
        except Exception as e:
            if not (self.quiet):
                logger.error(
                    f"Skipping {self.__class__.__name__} due to initialization "
                    f"failure!\n{e.__class__.__name__}: {e}"
                )
            self.failed()

        return self

    def run(self, hook=None):
        raise Exception(f"{COMMON_ERROR} Please override run method!")

    def post_run(self, hook=None):
        pass

    ###################################################
    #                     COMMONS                     #
    ###################################################

    def poke(self, msg: str = None):
        if self.is_running():
            return self.app
        raise Exception(
            msg or f"{self.__class__.__name__} is disabled or not yet configured!"
        )

    def is_ready(self):
        return self.state.is_ready() and self.app is None

    def is_running(self):
        return self.state.is_running() and not (self.app is None)

    def has_failed(self):
        return self.state.has_failed()

    def build_settings(self, hook) -> dict:
        try:
            self.manifest = self.build_manifest(hook)
            self.manifest_meta = {}
            if self.manifest:
                self.manifest_meta["__OLD_CONFIG__"] = self.manifest.pop(
                    "__OLD_CONFIG__", {}
                )
                self.manifest_meta["__UNSAFE_PARAPHRASE__"] = self.manifest.pop(
                    "__UNSAFE_PARAPHRASE__", ""
                )

            config = self.build_config(hook)
            self.enabled = config.pop("enabled", False)
            self.quiet = config.pop("quiet", False)
            self.config = config
        except Exception:
            logger.exception(f"Error loading settings for {self.__class__}")
            self.config = DEFAULT_CONFIG
            self.manifest = None
            self.manifest_meta = {}

    def build_config(self, hook) -> dict:
        return DEFAULT_CONFIG

    def build_manifest(self, hook) -> dict:
        pass

    # ------------------- DAEMON -------------------- #

    def spawn_daemon(self, **targets):
        for name, target in targets.items():
            dae: Process = self.daemon.get(name)
            if not dae or not dae.is_alive():
                process = Process(target=target, daemon=True)
                process.start()
                self.daemon[name] = process

    def terminate_daemon(self, *names):
        for name in names:
            dae: Process = self.daemon.pop(name, None)
            if not (dae is None) and dae.is_alive():
                logger.info(f"Terminating {name} ...")
                dae.terminate()

    ###################################################
    #                     CLEANER                     #
    ###################################################

    def reset(self, hook, start=True):
        self.app = None
        self.state = Ss.NOT_STARTED
        self.__init__(hook)

        if start:
            self.start(hook)

    def failed(self):
        self.app = None
        self.state = Ss.FAILED


class ProxyService(CommonService):
    def __init__(self):
        super().__init__(__class__)


class Kube:
    def __init__(self, in_cluster: bool, config: dict):
        if in_cluster:
            kubernetes_config.load_incluster_config()
        else:
            kubernetes_config.load_kube_config()

        self.client = ApiClient(config)
        self.core = CoreV1Api(config)
        self.api = AppsV1Api(self.client)
        self.auth = RbacAuthorizationV1Api(self.client)
        self.ping()
        self.defaults()

    def ping(self):
        res = self.client.call_api("/readyz", "GET")
        return res[1] == 200

    def create(self, api, namespace, conf):
        if api.startswith("ClusterRole"):
            self.create_apis[api](body=conf)
        else:
            self.create_apis[api](namespace=namespace, body=conf)

    def patch(self, api, name, namespace, conf):
        if api.startswith("ClusterRole"):
            self.patch_apis[api](name=name, body=conf)
        else:
            self.patch_apis[api](name=name, namespace=namespace, body=conf)

    def read(self, api: str, name: str, namespace: str = None):
        if api.startswith("ClusterRole"):
            return self.read_apis[api](name=name)
        else:
            return self.read_apis[api](name=name, namespace=namespace)

    def delete(self, api: str, name: str, namespace: str = None):
        if api.startswith("ClusterRole"):
            return self.delete_apis[api](name=name)
        else:
            return self.delete_apis[api](name=name, namespace=namespace)

    def is_running(self, name: str, namespace: str):
        try:
            return (
                self.core.list_namespaced_pod(
                    namespace=namespace, label_selector=f"pod={name}"
                )
                .items[0]
                .status.phase
                == "Running"
            )
        except Exception:
            return False

    def defaults(self):
        self.create_apis = {
            "Service": self.core.create_namespaced_service,
            "Deployment": self.api.create_namespaced_deployment,
            "ConfigMap": self.core.create_namespaced_config_map,
            "ServiceAccount": self.core.create_namespaced_service_account,
            "ClusterRole": self.auth.create_cluster_role,
            "ClusterRoleBinding": self.auth.create_cluster_role_binding,
            "Secret": self.core.create_namespaced_secret,
            "PersistentVolumeClaim": (
                self.core.create_namespaced_persistent_volume_claim
            ),
            "DaemonSet": self.api.create_namespaced_daemon_set,
        }
        self.patch_apis = {
            "Service": self.core.patch_namespaced_service,
            "Deployment": self.api.patch_namespaced_deployment,
            "ConfigMap": self.core.patch_namespaced_config_map,
            "ServiceAccount": self.core.patch_namespaced_service_account,
            "ClusterRole": self.auth.patch_cluster_role,
            "ClusterRoleBinding": self.auth.patch_cluster_role_binding,
            "Secret": self.core.patch_namespaced_secret,
            "PersistentVolumeClaim": (
                self.core.patch_namespaced_persistent_volume_claim
            ),
            "DaemonSet": self.api.patch_namespaced_daemon_set,
        }
        self.delete_apis = {
            "Service": self.core.delete_namespaced_service,
            "Deployment": self.api.delete_namespaced_deployment,
            "ConfigMap": self.core.delete_namespaced_config_map,
            "ServiceAccount": self.core.delete_namespaced_service_account,
            "ClusterRole": self.auth.delete_cluster_role,
            "ClusterRoleBinding": self.auth.delete_cluster_role_binding,
            "Secret": self.core.delete_namespaced_secret,
            "PersistentVolumeClaim": (
                self.core.delete_namespaced_persistent_volume_claim
            ),
            "DaemonSet": self.api.delete_namespaced_daemon_set,
        }
        self.read_apis = {
            "Service": self.core.read_namespaced_service,
            "Endpoints": self.core.read_namespaced_endpoints,
            "Deployment": self.api.read_namespaced_deployment,
            "ConfigMap": self.core.read_namespaced_config_map,
            "ServiceAccount": self.core.read_namespaced_service_account,
            "ClusterRole": self.auth.read_cluster_role,
            "ClusterRoleBinding": self.auth.read_cluster_role_binding,
            "Secret": self.core.read_namespaced_secret,
            "PersistentVolumeClaim": self.core.read_namespaced_persistent_volume_claim,
            "DaemonSet": self.api.read_namespaced_daemon_set,
        }


class JsOrc:
    def __init__(self, meta):
        self.automated = False
        self.meta = meta
        self.services = {}
        self.background = {}
        self.context = {}
        self.benchmark = {
            "jsorc": {"active": False, "requests": {}},
            "actions_optimizer": {"active": False, "requests": {}},
        }
        self.actions_history = {"active": False, "history": []}
        self.actions_calls = {}
        self.system_states = {"active": False, "states": []}
        self.actions_optimizer = ActionsOptimizer(
            benchmark=self.benchmark["actions_optimizer"],
            actions_history=self.actions_history,
            actions_calls=self.actions_calls,
        )

    ###################################################
    #                     BUILDER                     #
    ###################################################

    def build(self):
        try:
            hook = self.build_context("hook")
            config = hook.service_glob("META_CONFIG", META_CONFIG)
            if config.pop("automation", False):
                self.kubernetes = Kube(**config.pop("kubernetes", KUBERNETES_CONFIG))
                self.actions_optimizer.kube = self.kubernetes
                self.prometheus = self.meta.get_service("promon", hook)
                self.backoff_interval = config.pop("backoff_interval", 10)
                self.namespace = config.pop("namespace", "default")
                self.actions_optimizer.namespace = self.namespace
                self.keep_alive = config.pop(
                    "keep_alive", ["promon", "redis", "task", "mail"]
                )
                self.automated = True
            else:
                self.automated = False
        except Exception:
            self.automated = False

    def interval_check(self):
        hook = self.meta.build_hook()
        for svc in self.keep_alive:
            try:
                self.check(self.namespace, svc, hook)
            except Exception as e:
                logger.exception(
                    f"Error checking {svc} !\n" f"{e.__class__.__name__}: {e}"
                )
        self.optimize(jsorc_interval=self.backoff_interval)
        self.record_system_state()

    ###################################################
    #                   KUBERNETES                    #
    ###################################################

    def in_cluster(self):
        """
        Check if JSORC/Jaseci is running in a kubernetes cluster
        """
        try:
            if not hasattr(self, "kubernetes"):
                return False
            return self.kubernetes.ping()
        except ApiException as e:
            logger.info(f"Kubernetes cluster environment check failed: {e}")
            return False

    def create(self, kind: str, name: str, namespace: str, conf: dict):
        try:
            logger.info(f"Creating {kind} for `{name}` with namespace `{namespace}`")
            self.kubernetes.create(kind, namespace, conf)
        except ApiException as e:
            logger.error(
                f"Error creating {kind} for `{name}` with namespace `{namespace}` -- {e}"
            )

    def patch(self, kind: str, name: str, namespace: str, conf: dict):
        try:
            logger.info(f"Patching {kind} for `{name}` with namespace `{namespace}`")
            self.kubernetes.patch(kind, name, namespace, conf)
        except ApiException as e:
            logger.error(
                f"Error patching {kind} for `{name}` with namespace `{namespace}` -- {e}"
            )

    def read(self, kind: str, name: str, namespace: str):
        try:
            logger.info(f"Retrieving {kind} for `{name}` with namespace `{namespace}`")
            return self.kubernetes.read(kind, name, namespace=namespace)
        except ApiException as e:
            logger.error(
                f"Error retrieving {kind} for `{name}` with namespace `{namespace}` -- {e}"
            )
            return e

    def delete(self, kind: str, name: str, namespace: str):
        try:
            logger.info(f"Deleting {kind} for `{name}` with namespace `{namespace}`")
            return self.kubernetes.delete(kind, name, namespace=namespace)
        except ApiException as e:
            logger.error(
                f"Error deleting {kind} for `{name}` with namespace `{namespace}` -- {e}"
            )
            return e

    def check(self, namespace, svc_name, hook):

        svc = self.meta.get_service(svc_name, hook)

        if not svc.is_running():
            if svc.manifest:
                config_map = svc.manifest
                pod_name = ""
                old_config_map = deepcopy(svc.manifest_meta.get("__OLD_CONFIG__", {}))
                unsafe_paraphrase = svc.manifest_meta.get("__UNSAFE_PARAPHRASE__", "")
                for kind, confs in config_map.items():
                    for conf in confs:
                        name = conf["metadata"]["name"]
                        names = old_config_map.get(kind, [])
                        if name in names:
                            names.remove(name)

                        if kind == "Service":
                            pod_name = name
                        res = self.read(kind, name, namespace)
                        if hasattr(res, "status") and res.status == 404 and conf:
                            self.create(kind, name, namespace, conf)
                        elif not isinstance(res, ApiException) and res.metadata:
                            if res.metadata.labels:
                                config_version = res.metadata.labels.get(
                                    "config_version", 1
                                )
                            else:
                                config_version = 1

                            if config_version != conf.get("metadata").get(
                                "labels", {}
                            ).get("config_version", 1):
                                self.patch(kind, name, namespace, conf)

                    if (
                        old_config_map
                        and type(old_config_map) is dict
                        and kind in old_config_map
                        and name in old_config_map[kind]
                    ):
                        old_config_map.get(kind, []).remove(name)

                    for to_be_removed in old_config_map.get(kind, []):
                        res = self.read(kind, to_be_removed, namespace)
                        if not isinstance(res, ApiException) and res.metadata:
                            if (
                                kind not in UNSAFE_KINDS
                                or unsafe_paraphrase == UNSAFE_PARAPHRASE
                            ):
                                self.delete(kind, to_be_removed, namespace)
                            else:
                                logger.info(
                                    f"You don't have permission to delete `{kind}` for `{to_be_removed}` with namespace `{namespace}`!"
                                )

                if self.kubernetes.is_running(pod_name, namespace):
                    logger.info(
                        f"Pod state is running. Trying to Restart {svc_name}..."
                    )
                    svc.reset(hook)
            else:
                logger.info(f"Restarting {svc_name}...")
                svc.reset(hook)

    ###################################################
    #                 SERVICE HANDLER                 #
    ###################################################

    def add_service_builder(self, name, svc):
        if self.services.get(name):
            raise Exception(f"{name} already exists!")

        self.services[name] = svc

    def build_service(self, name, background, *args, **kwargs):

        svc = self.services.get(name)

        if not svc:
            logger.error(f"Service {name} is not yet set!")
            return None

        svc = svc(*args, **kwargs)

        if background:
            self.background[name] = svc

        return svc

    def get_service(self, name, *args, **kwargs):
        svc = self.background.get(name)

        if not svc:
            return self.build_service(name, True, *args, **kwargs)

        return svc

    ###################################################
    #                 CONTEXT HANDLER                 #
    ###################################################

    def add_context(self, name, cls, *args, **kwargs):
        self.context[name] = {"class": cls, "args": args, "kwargs": kwargs}

    def build_context(self, ctx, *args, **kwargs):
        ctx = self.context[ctx]
        return ctx["class"](*args, *ctx["args"], **kwargs, **ctx["kwargs"])

    ###################################################
    #                 ACTION MANAGER                  #
    ###################################################
    def load_action_config(self, name, config_mod):
        """
        Load the config for an action
        """
        return load_action_config(config_mod, name)

    def load_actions(self, name, mode):
        """
        Load an action as local, module or remote.
        """
        # Using module for local
        mode = "module" if mode == "local" else mode

        if mode == "module":
            self.actions_optimizer.load_action_module(name)
        elif mode == "remote":
            self.actions_optimizer.load_action_remote(name)

    def unload_actions(self, name, mode, retire_svc):
        """
        Unload an action
        """
        # We are using module for local
        mode = "module" if mode == "local" else mode
        if mode == "auto":
            res = self.actions_optimizer.unload_action_auto(name)
            if not res[0]:
                return res
            if retire_svc:
                self.retire_uservice(name)
            return res
        elif mode == "module":
            return self.actions_optimizer.unload_action_module(name)
        elif mode == "remote":
            res = self.actions_optimizer.unload_action_remote(name)
            if not res[0]:
                return res
            if retire_svc:
                self.retire_uservice(name)
            return res
        else:
            return (False, f"Unrecognized action mode {mode}.")

    def retire_uservice(self, name):
        """
        Retire a remote microservice for the action.
        """
        self.actions_optimizer.retire_remote(name)

    def get_actions_status(self, name=""):
        """
        Return the status of the action
        """
        return self.actions_optimizer.get_actions_status(name)

    def actions_tracking_start(self):
        """ """
        self.actions_history["active"] = True
        self.actions_history["history"] = [{"ts": time.time()}]
        self.actions_calls.clear()

    def actions_tracking_stop(self):
        """ """
        if not self.actions_history["active"]:
            return []

        self.actions_optimizer.summarize_action_calls()

        return self.actions_history["history"]

    def benchmark_start(self):
        """
        Put JSORC under benchmark mode.
        """
        self.benchmark["jsorc"]["active"] = True
        self.benchmark["jsorc"]["requests"] = {}
        self.benchmark["jsorc"]["start_ts"] = time.time()

    def state_tracking_start(self):
        """
        Ask JSORC to start tracking the state of the system as observed by JSORC on every interval.
        """
        self.system_states = {"active": True, "states": []}

    def state_tracking_stop(self):
        """
        Stop state tracking for JSORC
        """
        ret = self.system_states
        self.system_states = {"active": True, "states": []}
        return ret

    def state_tracking_report(self):
        """
        Return state tracking history so far
        """
        return self.system_states

    def record_system_state(self):
        """
        Record system state
        """
        if self.system_states["active"]:
            ts = int(time.time())
            prom_profile = self.prometheus.info(
                namespace=self.namespace,
                exclude_prom=True,
                timestamp=ts,
                duration=self.backoff_interval,
            )
            self.system_states["states"].append(
                {
                    "ts": ts,
                    "actions": self.get_actions_status(name=""),
                    "prometheus": prom_profile,
                }
            )

    def benchmark_stop(self, report):
        """
        Stop benchmark mode and report result during the benchmark period
        """
        if not self.benchmark["jsorc"]["active"]:
            return {}

        res = self.benchmark_report()
        self.benchmark["jsorc"]["requests"] = {}
        self.benchmark["jsorc"]["active"] = False

        if report:
            return res
        else:
            return {}

    def benchmark_report(self):
        """
        Summarize benchmark results and report.
        """
        summary = {}
        duration = time.time() - self.benchmark["jsorc"]["start_ts"]
        for request, data in self.benchmark["jsorc"]["requests"].items():
            summary[request] = {}
            all_reqs = []
            for req_name, times in data.items():
                if len(times) == 0:
                    continue
                all_reqs.extend(times)
                summary[request][req_name] = {
                    "throughput": len(times) / duration,
                    "average_latency": sum(times) / len(times) * 1000,
                    "50th_latency": np.percentile(times, 50) * 1000,
                    "90th_latency": np.percentile(times, 90) * 1000,
                    "95th_latency": np.percentile(times, 95) * 1000,
                    "99th_latency": np.percentile(times, 99) * 1000,
                }
            summary[request]["all"] = {
                "throughput": len(all_reqs) / duration,
                "average_latency": sum(all_reqs) / len(all_reqs) * 1000,
                "50th_latency": np.percentile(all_reqs, 50) * 1000,
                "90th_latency": np.percentile(all_reqs, 90) * 1000,
                "95th_latency": np.percentile(all_reqs, 95) * 1000,
                "99th_latency": np.percentile(all_reqs, 99) * 1000,
            }

        return summary

    def record_state(self):
        """
        Record the current state of the system observed by JSORC
        """

    def add_to_benchmark(self, request_type, request, request_time):
        """
        Add requests to benchmark performance tracking
        """
        for bm in self.benchmark.values():
            if request_type not in bm["requests"]:
                bm["requests"][request_type] = {"_default_": []}
            if request_type == "walker_run":
                walker_name = dict(request.data)["name"]
                if walker_name not in bm["requests"][request_type]:
                    bm["requests"][request_type][walker_name] = []
                bm["requests"][request_type][walker_name].append(request_time)
            else:
                bm["requests"][request_type]["_default_"].append(request_time)

    def set_action_policy(self, policy_name, policy_params):
        """
        Set an action optimizer policy
        """
        return self.actions_optimizer.set_action_policy(policy_name, policy_params)

    def get_action_policy(self):
        """
        Get the current action optimization policy
        """
        return self.actions_optimizer.get_action_policy()

    def pre_action_call_hook(self, *args):
        pass

    def post_action_call_hook(self, *args):
        action_name = args[0]
        action_time = args[1]
        if action_name not in self.actions_calls:
            self.actions_calls[action_name] = []

        self.actions_calls[action_name].append(action_time)

    def pre_request_hook(self, *args):
        pass

    def post_request_hook(self, *args):
        request_type = args[0]
        request = args[1]
        request_time = args[2]
        if self.benchmark["jsorc"]["active"]:
            self.add_to_benchmark(request_type, request, request_time)

    def optimize(self, jsorc_interval):
        self.actions_optimizer.run(jsorc_interval)


class MetaProperties:
    def __init__(self, cls):
        self.cls = cls

        if not hasattr(cls, "_app"):
            setattr(cls, "_app", None)
            setattr(cls, "_enabled", True)
            setattr(cls, "_state", Ss.NOT_STARTED)
            setattr(cls, "_quiet", False)
            setattr(cls, "_running_interval", 0)

    @property
    def app(self) -> JsOrc:
        return self.cls._app

    @app.setter
    def app(self, val: JsOrc):
        self.cls._app = val

    @property
    def state(self) -> Ss:
        return self.cls._state

    @state.setter
    def state(self, val: Ss):
        self.cls._state = val

    @property
    def enabled(self) -> bool:
        return self.cls._enabled

    @enabled.setter
    def enabled(self, val: bool):
        pass

    @property
    def quiet(self) -> bool:
        return self.cls._quiet

    @quiet.setter
    def quiet(self, val: bool):
        pass

    @property
    def running_interval(self) -> int:
        return self.cls._running_interval

    @running_interval.setter
    def running_interval(self, val: int):
        self.cls._running_interval = val



from jaseci.svc import CommonService
from .config import ELASTIC_CONFIG
from .manifest import ELASTIC_MANIFEST
from requests import get, post
from datetime import datetime
from copy import copy


#################################################
#                  ELASTIC APP                  #
#################################################


class ElasticService(CommonService):

    ###################################################
    #                     BUILDER                     #
    ###################################################

    def run(self, hook=None):
        self.app = Elastic(self.config)
        self.app.health("timeout=1s")

    ####################################################
    #                    OVERRIDDEN                    #
    ####################################################

    def build_config(self, hook) -> dict:
        return hook.service_glob("ELASTIC_CONFIG", ELASTIC_CONFIG)

    def build_manifest(self, hook) -> dict:
        return hook.service_glob("ELASTIC_MANIFEST", ELASTIC_MANIFEST)


class Elastic:
    def __init__(self, config):
        if not config.get("url"):
            raise Exception("URL is required!")
        self.url = config["url"]

        if not config.get("common_index"):
            raise Exception("Common Index is required!")
        self.common_index = config["common_index"]

        if not config.get("activity_index"):
            raise Exception("Activity Index is required!")
        self.activity_index = config["activity_index"]

        self.headers = {"Content-Type": "application/json"}

        if config["auth"]:
            self.headers["Authorization"] = config["auth"]

    def _get(self, url: str, json: dict = None):
        return get(f"{self.url}{url}", json=json, headers=self.headers).json()

    def _post(self, url: str, json: dict = None):
        return post(f"{self.url}{url}", json=json, headers=self.headers).json()

    def post(self, url: str, body: dict, index: str = "", suffix: str = ""):
        return self._post(f"/{index or self.common_index}{suffix}{url}", body)

    def post_act(self, url: str, body: dict, suffix: str = ""):
        return self.post(url, body, self.activity_index, suffix)

    def get(self, url: str, body: dict, index: str = "", suffix: str = ""):
        return self._get(f"/{index or self.common_index}{suffix}{url}", body)

    def get_act(self, url: str, body: dict, suffix: str = ""):
        return self.get(url, body, self.activity_index, suffix)

    def doc(self, log: dict, query: str = "", index: str = "", suffix: str = ""):
        return self.post(f"/_doc?{query}", log, index, suffix)

    def doc_activity(self, log: dict, query: str = "", suffix: str = ""):
        return self.doc(log, query, self.activity_index, suffix)

    def search(self, body: dict, query: str = "", index: str = "", suffix: str = ""):
        return self.get(f"/_search?{query}", body, index, suffix)

    def search_activity(self, body: dict, query: str = "", suffix: str = ""):
        return self.search(body, query, self.activity_index, suffix)

    def mapping(self, query: str = "", index: str = "", suffix: str = ""):
        return self.get(f"/_mapping?{query}", index, suffix)

    def mapping_activity(self, query: str = "", suffix: str = ""):
        return self.mapping(query, self.activity_index, suffix)

    def aliases(self, query: str = "pretty=true"):
        return self._get(f"/_aliases?{query}")

    def reindex(self, body: dict, query: str = "pretty"):
        return self._post(f"/_reindex?{query}", body)

    # standard methods
    def generate_from_meta(self, meta: dict, override: dict, action: str = None):
        scope = meta["scope"].local_scope
        interp = meta["interp"]

        node = interp.current_node
        walker = scope["visitor"]

        override_misc = override.get("misc")
        override["misc"] = {
            "report": copy(interp.report),
            "node": node.serialize(detailed=False),
        }

        if override_misc and isinstance(override_misc, dict):
            override["misc"].update(override_misc)

        master = meta["h"].get_obj(meta["m_id"], meta["m_id"]).master_self(True)

        headers = interp.request_context["headers"]
        if headers.get("Authorization"):
            del headers["Authorization"]

        activity = {
            "datetime": datetime.utcnow().isoformat(),
            "activity_action": action or walker.name.replace("_", " ").title(),
            "activity_type": walker.name,
            "activity_point": node.name,
            "walker_id": walker.jid,
            "node_id": node.jid,
            "master_id": meta["m_id"],
            "user": master.get("__meta__") or {"email": master["name"]},
            "request_context": interp.request_context,
            "data": walker.context,
        }

        activity.update(override)

        return activity

    def generate_from_request(self, request):
        headers = dict(request.headers)
        if headers.get("Authorization"):
            del headers["Authorization"]

        data = request.data.dict() if type(request.data) is not dict else request.data
        password = data.get("password")
        if password:
            data["password"] = len(password) * "*"

        user = request.user

        return {
            "datetime": datetime.utcnow().isoformat(),
            "activity_action": "User Manage",
            "activity_type": "user_manage",
            "master_id": user.master.urn,
            "user": {"email": user.email},
            "request_context": {
                "method": request.method,
                "headers": headers,
                "query": request.GET.dict(),
                "body": data,
            },
            "data": data,
        }

    def health(self, query: str = ""):
        return self._get(f"/_cluster/health?{query}")



KUBERNETES_CONFIG = {"in_cluster": True, "config": None}

META_CONFIG = {
    "automation": False,
    "backoff_interval": 10,
    "namespace": "default",
    "keep_alive": ["promon", "redis", "task", "mail", "elastic"],
    "kubernetes": KUBERNETES_CONFIG,
}



import os

DEFAULT_MSG = "Skipping scheduled walker!"
DEFAULT_URL = (
    f'redis://{os.getenv("REDIS_HOST", "localhost")}'
    f':{os.getenv("REDIS_PORT", "6379")}/{os.getenv("REDIS_DB", "1")}'
)

TASK_CONFIG = {
    "enabled": True,
    "quiet": True,
    "broker_url": DEFAULT_URL,
    "result_backend": DEFAULT_URL,
    "broker_connection_retry_on_startup": True,
    "task_track_started": True,
    "worker_redirect_stdouts": False,
}



import signal
from jaseci.utils.utils import logger
from jaseci.svc import (
    CommonService,
    JsOrc,
    MetaProperties,
    MailService,
    RedisService,
    TaskService,
    PrometheusService,
    ElasticService,
    ServiceState as Ss,
)


class MetaService(CommonService, MetaProperties):
    def __init__(self, run_svcs=True):
        self.run_svcs = run_svcs
        MetaProperties.__init__(self, __class__)

        self.start()

    ###################################################
    #                     BUILDER                     #
    ###################################################

    def run(self, hook=None):
        self.app = JsOrc(self)
        self.populate_context()
        self.populate_services()

    def post_run(self, hook=None):
        if self.run_svcs:
            self.app.build()
            if self.is_automated():
                logger.info("JsOrc is automated. Pushing interval check alarm...")
                self.push_interval(1)
            else:
                logger.info("JsOrc is not automated.")

    def push_interval(self, interval):
        if self.running_interval == 0:
            self.running_interval += 1
            signal.alarm(interval)
        else:
            logger.info("Reusing current running interval...")

    ###################################################
    #                    SERVICES                     #
    ###################################################

    def add_service_builder(self, name, svc):
        self.app.add_service_builder(name, svc)

    def build_service(self, name, background, *args, **kwargs):
        return self.app.build_service(name, background, *args, **kwargs)

    def get_service(self, name, *args, **kwargs):
        return self.app.get_service(name, *args, **kwargs)

    ###################################################
    #                    CONTEXT                      #
    ###################################################

    def add_context(self, ctx, cls, *args, **kwargs):
        self.app.add_context(ctx, cls, *args, **kwargs)

    def build_context(self, ctx, *args, **kwargs):
        return self.app.build_context(ctx, *args, **kwargs)

    def get_context(self, ctx):
        return self.app.context.get(ctx, {}).get("class")

    ###################################################
    #                     COMMON                      #
    ###################################################

    def is_automated(self):
        return self.is_running() and self.app and self.app.automated

    def in_cluster(self):
        return self.app.in_cluster()

    ###################################################
    #                     BUILDER                     #
    ###################################################

    def build_hook(self):
        h = self.build_context("hook")
        h.meta = self
        if self.run_svcs:
            h.promon = self.get_service("promon", h)
            h.redis = self.get_service("redis", h)
            h.task = self.get_service("task", h)
            h.mail = self.get_service("mail", h)
            h.elastic = self.get_service("elastic", h)

            if not self.is_automated():
                h.mail.start(h)
                h.redis.start(h)
                h.task.start(h)
                h.elastic.start(h)

        return h

    def build_master(self, *args, **kwargs):
        return self.__common("master", *args, **kwargs)

    def build_super_master(self, *args, **kwargs):
        return self.__common("super_master", *args, **kwargs)

    def __common(self, ctx, *args, **kwargs):
        if not kwargs.get("h", None):
            kwargs["h"] = self.build_hook()

        return self.build_context(ctx, *args, **kwargs)

    ###################################################
    #                   OVERRIDEN                     #
    ###################################################

    def reset(self, hook=None, start=True):
        self.terminate_daemon("jsorc")
        self.app = None
        self.state = Ss.NOT_STARTED
        self.__init__()

    def populate_context(self):
        from jaseci.hook import RedisHook
        from jaseci.element.master import Master
        from jaseci.element.super_master import SuperMaster

        self.add_context("hook", RedisHook)
        self.add_context("master", Master)
        self.add_context("super_master", SuperMaster)

    def populate_services(self):
        self.add_service_builder("redis", RedisService)
        self.add_service_builder("task", TaskService)
        self.add_service_builder("mail", MailService)
        self.add_service_builder("promon", PrometheusService)
        self.add_service_builder("elastic", ElasticService)


def interval_check(signum, frame):
    meta = MetaService()
    if meta.is_automated():
        meta.app.interval_check()
        logger.info(
            f"Backing off for {meta.app.backoff_interval} seconds before the next interval check..."
        )

        # wait interval_check to be finished before decrement
        meta.running_interval -= 1
        meta.push_interval(meta.app.backoff_interval)
    else:
        meta.running_interval -= 1


signal.signal(signal.SIGALRM, interval_check)



PROMON_MANIFEST = {
    "ServiceAccount": [
        {
            "apiVersion": "v1",
            "kind": "ServiceAccount",
            "metadata": {
                "labels": {
                    "helm.sh/chart": "kube-state-metrics-4.13.0",
                    "app.kubernetes.io/managed-by": "Helm",
                    "app.kubernetes.io/component": "metrics",
                    "app.kubernetes.io/part-of": "kube-state-metrics",
                    "app.kubernetes.io/name": "kube-state-metrics",
                    "app.kubernetes.io/instance": "jaseci-prometheus",
                    "app.kubernetes.io/version": "2.5.0",
                },
                "name": "jaseci-prometheus-kube-state-metrics",
            },
            "imagePullSecrets": [],
        },
        {
            "apiVersion": "v1",
            "kind": "ServiceAccount",
            "metadata": {
                "labels": {
                    "component": "alertmanager",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-alertmanager",
                "annotations": {},
            },
        },
        {
            "apiVersion": "v1",
            "kind": "ServiceAccount",
            "metadata": {
                "labels": {
                    "component": "node-exporter",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-node-exporter",
                "annotations": {},
            },
        },
        {
            "apiVersion": "v1",
            "kind": "ServiceAccount",
            "metadata": {
                "labels": {
                    "component": "pushgateway",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-pushgateway",
                "annotations": {},
            },
        },
        {
            "apiVersion": "v1",
            "kind": "ServiceAccount",
            "metadata": {
                "labels": {
                    "component": "server",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-server",
                "annotations": {},
            },
        },
    ],
    "ConfigMap": [
        {
            "apiVersion": "v1",
            "kind": "ConfigMap",
            "metadata": {
                "labels": {
                    "component": "alertmanager",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-alertmanager",
            },
            "data": {
                "allow-snippet-annotations": "false",
                "alertmanager.yml": "global: {}\nreceivers:\n- name: default-receiver\nroute:\n  group_interval: 5m\n  group_wait: 10s\n  receiver: default-receiver\n  repeat_interval: 3h\n",
            },
        },
        {
            "apiVersion": "v1",
            "kind": "ConfigMap",
            "metadata": {
                "labels": {
                    "component": "server",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-server",
            },
            "data": {
                "allow-snippet-annotations": "false",
                "alerting_rules.yml": "{}\n",
                "alerts": "{}\n",
                "prometheus.yml": 'global:\n  evaluation_interval: 1m\n  scrape_interval: 1m\n  scrape_timeout: 10s\nrule_files:\n- /etc/config/recording_rules.yml\n- /etc/config/alerting_rules.yml\n- /etc/config/rules\n- /etc/config/alerts\nscrape_configs:\n- job_name: prometheus\n  static_configs:\n  - targets:\n    - localhost:9090\n- bearer_token_file: /var/run/secrets/kubernetes.io/serviceaccount/token\n  job_name: kubernetes-apiservers\n  kubernetes_sd_configs:\n  - role: endpoints\n  relabel_configs:\n  - action: keep\n    regex: default;kubernetes;https\n    source_labels:\n    - __meta_kubernetes_namespace\n    - __meta_kubernetes_service_name\n    - __meta_kubernetes_endpoint_port_name\n  scheme: https\n  tls_config:\n    ca_file: /var/run/secrets/kubernetes.io/serviceaccount/ca.crt\n    insecure_skip_verify: true\n- bearer_token_file: /var/run/secrets/kubernetes.io/serviceaccount/token\n  job_name: kubernetes-nodes\n  kubernetes_sd_configs:\n  - role: node\n  relabel_configs:\n  - action: labelmap\n    regex: __meta_kubernetes_node_label_(.+)\n  - replacement: kubernetes.default.svc:443\n    target_label: __address__\n  - regex: (.+)\n    replacement: /api/v1/nodes/$1/proxy/metrics\n    source_labels:\n    - __meta_kubernetes_node_name\n    target_label: __metrics_path__\n  scheme: https\n  tls_config:\n    ca_file: /var/run/secrets/kubernetes.io/serviceaccount/ca.crt\n    insecure_skip_verify: true\n- bearer_token_file: /var/run/secrets/kubernetes.io/serviceaccount/token\n  job_name: kubernetes-nodes-cadvisor\n  kubernetes_sd_configs:\n  - role: node\n  relabel_configs:\n  - action: labelmap\n    regex: __meta_kubernetes_node_label_(.+)\n  - replacement: kubernetes.default.svc:443\n    target_label: __address__\n  - regex: (.+)\n    replacement: /api/v1/nodes/$1/proxy/metrics/cadvisor\n    source_labels:\n    - __meta_kubernetes_node_name\n    target_label: __metrics_path__\n  scheme: https\n  tls_config:\n    ca_file: /var/run/secrets/kubernetes.io/serviceaccount/ca.crt\n    insecure_skip_verify: true\n- honor_labels: true\n  job_name: kubernetes-service-endpoints\n  kubernetes_sd_configs:\n  - role: endpoints\n  relabel_configs:\n  - action: keep\n    regex: true\n    source_labels:\n    - __meta_kubernetes_service_annotation_prometheus_io_scrape\n  - action: drop\n    regex: true\n    source_labels:\n    - __meta_kubernetes_service_annotation_prometheus_io_scrape_slow\n  - action: replace\n    regex: (https?)\n    source_labels:\n    - __meta_kubernetes_service_annotation_prometheus_io_scheme\n    target_label: __scheme__\n  - action: replace\n    regex: (.+)\n    source_labels:\n    - __meta_kubernetes_service_annotation_prometheus_io_path\n    target_label: __metrics_path__\n  - action: replace\n    regex: (.+?)(?::\\d+)?;(\\d+)\n    replacement: $1:$2\n    source_labels:\n    - __address__\n    - __meta_kubernetes_service_annotation_prometheus_io_port\n    target_label: __address__\n  - action: labelmap\n    regex: __meta_kubernetes_service_annotation_prometheus_io_param_(.+)\n    replacement: __param_$1\n  - action: labelmap\n    regex: __meta_kubernetes_service_label_(.+)\n  - action: replace\n    source_labels:\n    - __meta_kubernetes_namespace\n    target_label: namespace\n  - action: replace\n    source_labels:\n    - __meta_kubernetes_service_name\n    target_label: service\n  - action: replace\n    source_labels:\n    - __meta_kubernetes_pod_node_name\n    target_label: node\n- honor_labels: true\n  job_name: kubernetes-service-endpoints-slow\n  kubernetes_sd_configs:\n  - role: endpoints\n  relabel_configs:\n  - action: keep\n    regex: true\n    source_labels:\n    - __meta_kubernetes_service_annotation_prometheus_io_scrape_slow\n  - action: replace\n    regex: (https?)\n    source_labels:\n    - __meta_kubernetes_service_annotation_prometheus_io_scheme\n    target_label: __scheme__\n  - action: replace\n    regex: (.+)\n    source_labels:\n    - __meta_kubernetes_service_annotation_prometheus_io_path\n    target_label: __metrics_path__\n  - action: replace\n    regex: (.+?)(?::\\d+)?;(\\d+)\n    replacement: $1:$2\n    source_labels:\n    - __address__\n    - __meta_kubernetes_service_annotation_prometheus_io_port\n    target_label: __address__\n  - action: labelmap\n    regex: __meta_kubernetes_service_annotation_prometheus_io_param_(.+)\n    replacement: __param_$1\n  - action: labelmap\n    regex: __meta_kubernetes_service_label_(.+)\n  - action: replace\n    source_labels:\n    - __meta_kubernetes_namespace\n    target_label: namespace\n  - action: replace\n    source_labels:\n    - __meta_kubernetes_service_name\n    target_label: service\n  - action: replace\n    source_labels:\n    - __meta_kubernetes_pod_node_name\n    target_label: node\n  scrape_interval: 5m\n  scrape_timeout: 30s\n- honor_labels: true\n  job_name: prometheus-pushgateway\n  kubernetes_sd_configs:\n  - role: service\n  relabel_configs:\n  - action: keep\n    regex: pushgateway\n    source_labels:\n    - __meta_kubernetes_service_annotation_prometheus_io_probe\n- honor_labels: true\n  job_name: kubernetes-services\n  kubernetes_sd_configs:\n  - role: service\n  metrics_path: /probe\n  params:\n    module:\n    - http_2xx\n  relabel_configs:\n  - action: keep\n    regex: true\n    source_labels:\n    - __meta_kubernetes_service_annotation_prometheus_io_probe\n  - source_labels:\n    - __address__\n    target_label: __param_target\n  - replacement: blackbox\n    target_label: __address__\n  - source_labels:\n    - __param_target\n    target_label: instance\n  - action: labelmap\n    regex: __meta_kubernetes_service_label_(.+)\n  - source_labels:\n    - __meta_kubernetes_namespace\n    target_label: namespace\n  - source_labels:\n    - __meta_kubernetes_service_name\n    target_label: service\n- honor_labels: true\n  job_name: kubernetes-pods\n  kubernetes_sd_configs:\n  - role: pod\n  relabel_configs:\n  - action: keep\n    regex: true\n    source_labels:\n    - __meta_kubernetes_pod_annotation_prometheus_io_scrape\n  - action: drop\n    regex: true\n    source_labels:\n    - __meta_kubernetes_pod_annotation_prometheus_io_scrape_slow\n  - action: replace\n    regex: (https?)\n    source_labels:\n    - __meta_kubernetes_pod_annotation_prometheus_io_scheme\n    target_label: __scheme__\n  - action: replace\n    regex: (.+)\n    source_labels:\n    - __meta_kubernetes_pod_annotation_prometheus_io_path\n    target_label: __metrics_path__\n  - action: replace\n    regex: (.+?)(?::\\d+)?;(\\d+)\n    replacement: $1:$2\n    source_labels:\n    - __address__\n    - __meta_kubernetes_pod_annotation_prometheus_io_port\n    target_label: __address__\n  - action: labelmap\n    regex: __meta_kubernetes_pod_annotation_prometheus_io_param_(.+)\n    replacement: __param_$1\n  - action: labelmap\n    regex: __meta_kubernetes_pod_label_(.+)\n  - action: replace\n    source_labels:\n    - __meta_kubernetes_namespace\n    target_label: namespace\n  - action: replace\n    source_labels:\n    - __meta_kubernetes_pod_name\n    target_label: pod\n  - action: drop\n    regex: Pending|Succeeded|Failed|Completed\n    source_labels:\n    - __meta_kubernetes_pod_phase\n- honor_labels: true\n  job_name: kubernetes-pods-slow\n  kubernetes_sd_configs:\n  - role: pod\n  relabel_configs:\n  - action: keep\n    regex: true\n    source_labels:\n    - __meta_kubernetes_pod_annotation_prometheus_io_scrape_slow\n  - action: replace\n    regex: (https?)\n    source_labels:\n    - __meta_kubernetes_pod_annotation_prometheus_io_scheme\n    target_label: __scheme__\n  - action: replace\n    regex: (.+)\n    source_labels:\n    - __meta_kubernetes_pod_annotation_prometheus_io_path\n    target_label: __metrics_path__\n  - action: replace\n    regex: (.+?)(?::\\d+)?;(\\d+)\n    replacement: $1:$2\n    source_labels:\n    - __address__\n    - __meta_kubernetes_pod_annotation_prometheus_io_port\n    target_label: __address__\n  - action: labelmap\n    regex: __meta_kubernetes_pod_annotation_prometheus_io_param_(.+)\n    replacement: __param_$1\n  - action: labelmap\n    regex: __meta_kubernetes_pod_label_(.+)\n  - action: replace\n    source_labels:\n    - __meta_kubernetes_namespace\n    target_label: namespace\n  - action: replace\n    source_labels:\n    - __meta_kubernetes_pod_name\n    target_label: pod\n  - action: drop\n    regex: Pending|Succeeded|Failed|Completed\n    source_labels:\n    - __meta_kubernetes_pod_phase\n  scrape_interval: 5m\n  scrape_timeout: 30s\nalerting:\n  alertmanagers:\n  - kubernetes_sd_configs:\n      - role: pod\n    tls_config:\n      ca_file: /var/run/secrets/kubernetes.io/serviceaccount/ca.crt\n    bearer_token_file: /var/run/secrets/kubernetes.io/serviceaccount/token\n    relabel_configs:\n    - source_labels: [__meta_kubernetes_namespace]\n      regex: default\n      action: keep\n    - source_labels: [__meta_kubernetes_pod_label_app]\n      regex: prometheus\n      action: keep\n    - source_labels: [__meta_kubernetes_pod_label_component]\n      regex: alertmanager\n      action: keep\n    - source_labels: [__meta_kubernetes_pod_annotation_prometheus_io_probe]\n      regex: .*\n      action: keep\n    - source_labels: [__meta_kubernetes_pod_container_port_number]\n      regex: "9093"\n      action: keep\n',
                "recording_rules.yml": "{}\n",
                "rules": "{}\n",
            },
        },
    ],
    "PersistentVolumeClaim": [
        {
            "apiVersion": "v1",
            "kind": "PersistentVolumeClaim",
            "metadata": {
                "labels": {
                    "component": "alertmanager",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-alertmanager",
            },
            "spec": {
                "accessModes": ["ReadWriteOnce"],
                "resources": {"requests": {"storage": "2Gi"}},
            },
        },
        {
            "apiVersion": "v1",
            "kind": "PersistentVolumeClaim",
            "metadata": {
                "labels": {
                    "component": "server",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-server",
            },
            "spec": {
                "accessModes": ["ReadWriteOnce"],
                "resources": {"requests": {"storage": "8Gi"}},
            },
        },
    ],
    "ClusterRole": [
        {
            "apiVersion": "rbac.authorization.k8s.io/v1",
            "kind": "ClusterRole",
            "metadata": {
                "labels": {
                    "helm.sh/chart": "kube-state-metrics-4.13.0",
                    "app.kubernetes.io/managed-by": "Helm",
                    "app.kubernetes.io/component": "metrics",
                    "app.kubernetes.io/part-of": "kube-state-metrics",
                    "app.kubernetes.io/name": "kube-state-metrics",
                    "app.kubernetes.io/instance": "jaseci-prometheus",
                    "app.kubernetes.io/version": "2.5.0",
                },
                "name": "jaseci-prometheus-kube-state-metrics",
            },
            "rules": [
                {
                    "apiGroups": ["certificates.k8s.io"],
                    "resources": ["certificatesigningrequests"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": [""],
                    "resources": ["configmaps"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["batch"],
                    "resources": ["cronjobs"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["extensions", "apps"],
                    "resources": ["daemonsets"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["extensions", "apps"],
                    "resources": ["deployments"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": [""],
                    "resources": ["endpoints"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["autoscaling"],
                    "resources": ["horizontalpodautoscalers"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["extensions", "networking.k8s.io"],
                    "resources": ["ingresses"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["batch"],
                    "resources": ["jobs"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": [""],
                    "resources": ["limitranges"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["admissionregistration.k8s.io"],
                    "resources": ["mutatingwebhookconfigurations"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": [""],
                    "resources": ["namespaces"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["networking.k8s.io"],
                    "resources": ["networkpolicies"],
                    "verbs": ["list", "watch"],
                },
                {"apiGroups": [""], "resources": ["nodes"], "verbs": ["list", "watch"]},
                {
                    "apiGroups": [""],
                    "resources": ["persistentvolumeclaims"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": [""],
                    "resources": ["persistentvolumes"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["policy"],
                    "resources": ["poddisruptionbudgets"],
                    "verbs": ["list", "watch"],
                },
                {"apiGroups": [""], "resources": ["pods"], "verbs": ["list", "watch"]},
                {
                    "apiGroups": ["extensions", "apps"],
                    "resources": ["replicasets"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": [""],
                    "resources": ["replicationcontrollers"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": [""],
                    "resources": ["resourcequotas"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": [""],
                    "resources": ["secrets"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": [""],
                    "resources": ["services"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["apps"],
                    "resources": ["statefulsets"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["storage.k8s.io"],
                    "resources": ["storageclasses"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["admissionregistration.k8s.io"],
                    "resources": ["validatingwebhookconfigurations"],
                    "verbs": ["list", "watch"],
                },
                {
                    "apiGroups": ["storage.k8s.io"],
                    "resources": ["volumeattachments"],
                    "verbs": ["list", "watch"],
                },
            ],
        },
        {
            "apiVersion": "rbac.authorization.k8s.io/v1",
            "kind": "ClusterRole",
            "metadata": {
                "labels": {
                    "component": "alertmanager",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-alertmanager",
            },
            "rules": [],
        },
        {
            "apiVersion": "rbac.authorization.k8s.io/v1",
            "kind": "ClusterRole",
            "metadata": {
                "labels": {
                    "component": "pushgateway",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-pushgateway",
            },
            "rules": [],
        },
        {
            "apiVersion": "rbac.authorization.k8s.io/v1",
            "kind": "ClusterRole",
            "metadata": {
                "labels": {
                    "component": "server",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-server",
            },
            "rules": [
                {
                    "apiGroups": [""],
                    "resources": [
                        "nodes",
                        "nodes/proxy",
                        "nodes/metrics",
                        "services",
                        "endpoints",
                        "pods",
                        "ingresses",
                        "configmaps",
                    ],
                    "verbs": ["get", "list", "watch"],
                },
                {
                    "apiGroups": ["extensions", "networking.k8s.io"],
                    "resources": ["ingresses/status", "ingresses"],
                    "verbs": ["get", "list", "watch"],
                },
                {"nonResourceURLs": ["/metrics"], "verbs": ["get"]},
            ],
        },
    ],
    "ClusterRoleBinding": [
        {
            "apiVersion": "rbac.authorization.k8s.io/v1",
            "kind": "ClusterRoleBinding",
            "metadata": {
                "labels": {
                    "helm.sh/chart": "kube-state-metrics-4.13.0",
                    "app.kubernetes.io/managed-by": "Helm",
                    "app.kubernetes.io/component": "metrics",
                    "app.kubernetes.io/part-of": "kube-state-metrics",
                    "app.kubernetes.io/name": "kube-state-metrics",
                    "app.kubernetes.io/instance": "jaseci-prometheus",
                    "app.kubernetes.io/version": "2.5.0",
                },
                "name": "jaseci-prometheus-kube-state-metrics",
            },
            "roleRef": {
                "apiGroup": "rbac.authorization.k8s.io",
                "kind": "ClusterRole",
                "name": "jaseci-prometheus-kube-state-metrics",
            },
            "subjects": [
                {
                    "kind": "ServiceAccount",
                    "name": "jaseci-prometheus-kube-state-metrics",
                    "namespace": "default",
                }
            ],
        },
        {
            "apiVersion": "rbac.authorization.k8s.io/v1",
            "kind": "ClusterRoleBinding",
            "metadata": {
                "labels": {
                    "component": "alertmanager",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-alertmanager",
            },
            "subjects": [
                {
                    "kind": "ServiceAccount",
                    "name": "jaseci-prometheus-alertmanager",
                    "namespace": "default",
                }
            ],
            "roleRef": {
                "apiGroup": "rbac.authorization.k8s.io",
                "kind": "ClusterRole",
                "name": "jaseci-prometheus-alertmanager",
            },
        },
        {
            "apiVersion": "rbac.authorization.k8s.io/v1",
            "kind": "ClusterRoleBinding",
            "metadata": {
                "labels": {
                    "component": "pushgateway",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-pushgateway",
            },
            "subjects": [
                {
                    "kind": "ServiceAccount",
                    "name": "jaseci-prometheus-pushgateway",
                    "namespace": "default",
                }
            ],
            "roleRef": {
                "apiGroup": "rbac.authorization.k8s.io",
                "kind": "ClusterRole",
                "name": "jaseci-prometheus-pushgateway",
            },
        },
        {
            "apiVersion": "rbac.authorization.k8s.io/v1",
            "kind": "ClusterRoleBinding",
            "metadata": {
                "labels": {
                    "component": "server",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-server",
            },
            "subjects": [
                {
                    "kind": "ServiceAccount",
                    "name": "jaseci-prometheus-server",
                    "namespace": "default",
                }
            ],
            "roleRef": {
                "apiGroup": "rbac.authorization.k8s.io",
                "kind": "ClusterRole",
                "name": "jaseci-prometheus-server",
            },
        },
    ],
    "Service": [
        {
            "apiVersion": "v1",
            "kind": "Service",
            "metadata": {
                "name": "jaseci-prometheus-kube-state-metrics",
                "labels": {
                    "helm.sh/chart": "kube-state-metrics-4.13.0",
                    "app.kubernetes.io/managed-by": "Helm",
                    "app.kubernetes.io/component": "metrics",
                    "app.kubernetes.io/part-of": "kube-state-metrics",
                    "app.kubernetes.io/name": "kube-state-metrics",
                    "app.kubernetes.io/instance": "jaseci-prometheus",
                    "app.kubernetes.io/version": "2.5.0",
                },
                "annotations": {"prometheus.io/scrape": "true"},
            },
            "spec": {
                "type": "ClusterIP",
                "ports": [
                    {
                        "name": "http",
                        "protocol": "TCP",
                        "port": 8080,
                        "targetPort": 8080,
                    }
                ],
                "selector": {
                    "app.kubernetes.io/name": "kube-state-metrics",
                    "app.kubernetes.io/instance": "jaseci-prometheus",
                },
            },
        },
        {
            "apiVersion": "v1",
            "kind": "Service",
            "metadata": {
                "labels": {
                    "component": "alertmanager",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-alertmanager",
            },
            "spec": {
                "ports": [
                    {"name": "http", "port": 80, "protocol": "TCP", "targetPort": 9093}
                ],
                "selector": {
                    "component": "alertmanager",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                },
                "sessionAffinity": "None",
                "type": "ClusterIP",
            },
        },
        {
            "apiVersion": "v1",
            "kind": "Service",
            "metadata": {
                "annotations": {"prometheus.io/scrape": "true"},
                "labels": {
                    "component": "node-exporter",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-node-exporter",
            },
            "spec": {
                "ports": [
                    {
                        "name": "metrics",
                        "port": 9100,
                        "protocol": "TCP",
                        "targetPort": 9100,
                    }
                ],
                "selector": {
                    "component": "node-exporter",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                },
                "type": "ClusterIP",
            },
        },
        {
            "apiVersion": "v1",
            "kind": "Service",
            "metadata": {
                "annotations": {"prometheus.io/probe": "pushgateway"},
                "labels": {
                    "component": "pushgateway",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-pushgateway",
            },
            "spec": {
                "ports": [
                    {
                        "name": "http",
                        "port": 9091,
                        "protocol": "TCP",
                        "targetPort": 9091,
                    }
                ],
                "selector": {
                    "component": "pushgateway",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                },
                "type": "ClusterIP",
            },
        },
        {
            "apiVersion": "v1",
            "kind": "Service",
            "metadata": {
                "labels": {
                    "component": "server",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-server",
            },
            "spec": {
                "ports": [
                    {
                        "name": "http",
                        "port": 9090,
                        "protocol": "TCP",
                        "targetPort": 9090,
                    }
                ],
                "selector": {
                    "component": "server",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                },
                "sessionAffinity": "None",
                "type": "ClusterIP",
            },
        },
    ],
    "DaemonSet": [
        {
            "apiVersion": "apps/v1",
            "kind": "DaemonSet",
            "metadata": {
                "labels": {
                    "component": "node-exporter",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-node-exporter",
            },
            "spec": {
                "selector": {
                    "matchLabels": {
                        "component": "node-exporter",
                        "app": "prometheus",
                        "release": "jaseci-prometheus",
                    }
                },
                "updateStrategy": {"type": "RollingUpdate"},
                "template": {
                    "metadata": {
                        "labels": {
                            "component": "node-exporter",
                            "app": "prometheus",
                            "release": "jaseci-prometheus",
                            "chart": "prometheus-15.10.5",
                            "heritage": "Helm",
                        }
                    },
                    "spec": {
                        "serviceAccountName": "jaseci-prometheus-node-exporter",
                        "containers": [
                            {
                                "name": "prometheus-node-exporter",
                                "image": "quay.io/prometheus/node-exporter:v1.3.1",
                                "imagePullPolicy": "IfNotPresent",
                                "args": [
                                    "--path.procfs=/host/proc",
                                    "--path.sysfs=/host/sys",
                                    "--web.listen-address=:9100",
                                ],
                                "ports": [
                                    {
                                        "name": "metrics",
                                        "containerPort": 9100,
                                        "hostPort": 9100,
                                    }
                                ],
                                "resources": {},
                                "securityContext": {"allowPrivilegeEscalation": False},
                                "volumeMounts": [
                                    {
                                        "name": "proc",
                                        "mountPath": "/host/proc",
                                        "readOnly": True,
                                    },
                                    {
                                        "name": "sys",
                                        "mountPath": "/host/sys",
                                        "readOnly": True,
                                    },
                                ],
                            }
                        ],
                        "hostNetwork": True,
                        "hostPID": True,
                        "securityContext": {
                            "fsGroup": 65534,
                            "runAsGroup": 65534,
                            "runAsNonRoot": True,
                            "runAsUser": 65534,
                        },
                        "volumes": [
                            {"name": "proc", "hostPath": {"path": "/proc"}},
                            {"name": "sys", "hostPath": {"path": "/sys"}},
                        ],
                    },
                },
            },
        }
    ],
    "Deployment": [
        {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": {
                "name": "jaseci-prometheus-kube-state-metrics",
                "labels": {
                    "helm.sh/chart": "kube-state-metrics-4.13.0",
                    "app.kubernetes.io/managed-by": "Helm",
                    "app.kubernetes.io/component": "metrics",
                    "app.kubernetes.io/part-of": "kube-state-metrics",
                    "app.kubernetes.io/name": "kube-state-metrics",
                    "app.kubernetes.io/instance": "jaseci-prometheus",
                    "app.kubernetes.io/version": "2.5.0",
                },
            },
            "spec": {
                "selector": {
                    "matchLabels": {
                        "app.kubernetes.io/name": "kube-state-metrics",
                        "app.kubernetes.io/instance": "jaseci-prometheus",
                    }
                },
                "replicas": 1,
                "template": {
                    "metadata": {
                        "labels": {
                            "helm.sh/chart": "kube-state-metrics-4.13.0",
                            "app.kubernetes.io/managed-by": "Helm",
                            "app.kubernetes.io/component": "metrics",
                            "app.kubernetes.io/part-of": "kube-state-metrics",
                            "app.kubernetes.io/name": "kube-state-metrics",
                            "app.kubernetes.io/instance": "jaseci-prometheus",
                            "app.kubernetes.io/version": "2.5.0",
                        }
                    },
                    "spec": {
                        "hostNetwork": False,
                        "serviceAccountName": "jaseci-prometheus-kube-state-metrics",
                        "securityContext": {
                            "fsGroup": 65534,
                            "runAsGroup": 65534,
                            "runAsUser": 65534,
                        },
                        "containers": [
                            {
                                "name": "kube-state-metrics",
                                "args": [
                                    "--port=8080",
                                    "--resources=certificatesigningrequests,configmaps,cronjobs,daemonsets,deployments,endpoints,horizontalpodautoscalers,ingresses,jobs,limitranges,mutatingwebhookconfigurations,namespaces,networkpolicies,nodes,persistentvolumeclaims,persistentvolumes,poddisruptionbudgets,pods,replicasets,replicationcontrollers,resourcequotas,secrets,services,statefulsets,storageclasses,validatingwebhookconfigurations,volumeattachments",
                                    "--telemetry-port=8081",
                                ],
                                "imagePullPolicy": "IfNotPresent",
                                "image": "registry.k8s.io/kube-state-metrics/kube-state-metrics:v2.5.0",
                                "ports": [{"containerPort": 8080, "name": "http"}],
                                "livenessProbe": {
                                    "httpGet": {"path": "/healthz", "port": 8080},
                                    "initialDelaySeconds": 5,
                                    "timeoutSeconds": 5,
                                },
                                "readinessProbe": {
                                    "httpGet": {"path": "/", "port": 8080},
                                    "initialDelaySeconds": 5,
                                    "timeoutSeconds": 5,
                                },
                            }
                        ],
                    },
                },
            },
        },
        {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": {
                "labels": {
                    "component": "alertmanager",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-alertmanager",
            },
            "spec": {
                "selector": {
                    "matchLabels": {
                        "component": "alertmanager",
                        "app": "prometheus",
                        "release": "jaseci-prometheus",
                    }
                },
                "replicas": 1,
                "template": {
                    "metadata": {
                        "labels": {
                            "component": "alertmanager",
                            "app": "prometheus",
                            "release": "jaseci-prometheus",
                            "chart": "prometheus-15.10.5",
                            "heritage": "Helm",
                        }
                    },
                    "spec": {
                        "serviceAccountName": "jaseci-prometheus-alertmanager",
                        "containers": [
                            {
                                "name": "prometheus-alertmanager",
                                "image": "quay.io/prometheus/alertmanager:v0.24.0",
                                "imagePullPolicy": "IfNotPresent",
                                "securityContext": {},
                                "env": [
                                    {
                                        "name": "POD_IP",
                                        "valueFrom": {
                                            "fieldRef": {
                                                "apiVersion": "v1",
                                                "fieldPath": "status.podIP",
                                            }
                                        },
                                    }
                                ],
                                "args": [
                                    "--config.file=/etc/config/alertmanager.yml",
                                    "--storage.path=/data",
                                    "--cluster.listen-address=",
                                    "--web.external-url=http://localhost:9093",
                                ],
                                "ports": [{"containerPort": 9093}],
                                "readinessProbe": {
                                    "httpGet": {"path": "/-/ready", "port": 9093},
                                    "initialDelaySeconds": 30,
                                    "timeoutSeconds": 30,
                                },
                                "resources": {},
                                "volumeMounts": [
                                    {
                                        "name": "config-volume",
                                        "mountPath": "/etc/config",
                                    },
                                    {
                                        "name": "storage-volume",
                                        "mountPath": "/data",
                                        "subPath": "",
                                    },
                                ],
                            },
                            {
                                "name": "prometheus-alertmanager-configmap-reload",
                                "image": "jimmidyson/configmap-reload:v0.5.0",
                                "imagePullPolicy": "IfNotPresent",
                                "securityContext": {},
                                "args": [
                                    "--volume-dir=/etc/config",
                                    "--webhook-url=http://127.0.0.1:9093/-/reload",
                                ],
                                "resources": {},
                                "volumeMounts": [
                                    {
                                        "name": "config-volume",
                                        "mountPath": "/etc/config",
                                        "readOnly": True,
                                    }
                                ],
                            },
                        ],
                        "securityContext": {
                            "fsGroup": 65534,
                            "runAsGroup": 65534,
                            "runAsNonRoot": True,
                            "runAsUser": 65534,
                        },
                        "volumes": [
                            {
                                "name": "config-volume",
                                "configMap": {"name": "jaseci-prometheus-alertmanager"},
                            },
                            {
                                "name": "storage-volume",
                                "persistentVolumeClaim": {
                                    "claimName": "jaseci-prometheus-alertmanager"
                                },
                            },
                        ],
                    },
                },
            },
        },
        {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": {
                "labels": {
                    "component": "pushgateway",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                },
                "name": "jaseci-prometheus-pushgateway",
            },
            "spec": {
                "selector": {
                    "matchLabels": {
                        "component": "pushgateway",
                        "app": "prometheus",
                        "release": "jaseci-prometheus",
                    }
                },
                "replicas": 1,
                "template": {
                    "metadata": {
                        "labels": {
                            "component": "pushgateway",
                            "app": "prometheus",
                            "release": "jaseci-prometheus",
                            "chart": "prometheus-15.10.5",
                            "heritage": "Helm",
                        }
                    },
                    "spec": {
                        "serviceAccountName": "jaseci-prometheus-pushgateway",
                        "containers": [
                            {
                                "name": "prometheus-pushgateway",
                                "image": "prom/pushgateway:v1.4.3",
                                "imagePullPolicy": "IfNotPresent",
                                "securityContext": {},
                                "args": None,
                                "ports": [{"containerPort": 9091}],
                                "livenessProbe": {
                                    "httpGet": {"path": "/-/healthy", "port": 9091},
                                    "initialDelaySeconds": 10,
                                    "timeoutSeconds": 10,
                                },
                                "readinessProbe": {
                                    "httpGet": {"path": "/-/ready", "port": 9091},
                                    "initialDelaySeconds": 10,
                                    "timeoutSeconds": 10,
                                },
                                "resources": {},
                            }
                        ],
                        "securityContext": {"runAsNonRoot": True, "runAsUser": 65534},
                    },
                },
            },
        },
        {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": {
                "labels": {
                    "component": "server",
                    "app": "prometheus",
                    "release": "jaseci-prometheus",
                    "chart": "prometheus-15.10.5",
                    "heritage": "Helm",
                    "pod": "jaseci-prometheus-server",
                },
                "name": "jaseci-prometheus-server",
            },
            "spec": {
                "selector": {
                    "matchLabels": {
                        "component": "server",
                        "app": "prometheus",
                        "release": "jaseci-prometheus",
                    }
                },
                "replicas": 1,
                "template": {
                    "metadata": {
                        "labels": {
                            "component": "server",
                            "app": "prometheus",
                            "release": "jaseci-prometheus",
                            "chart": "prometheus-15.10.5",
                            "heritage": "Helm",
                            "pod": "jaseci-prometheus-server",
                        }
                    },
                    "spec": {
                        "enableServiceLinks": True,
                        "serviceAccountName": "jaseci-prometheus-server",
                        "containers": [
                            {
                                "name": "prometheus-server-configmap-reload",
                                "image": "jimmidyson/configmap-reload:v0.5.0",
                                "imagePullPolicy": "IfNotPresent",
                                "securityContext": {},
                                "args": [
                                    "--volume-dir=/etc/config",
                                    "--webhook-url=http://127.0.0.1:9090/-/reload",
                                ],
                                "resources": {},
                                "volumeMounts": [
                                    {
                                        "name": "config-volume",
                                        "mountPath": "/etc/config",
                                        "readOnly": True,
                                    }
                                ],
                            },
                            {
                                "name": "prometheus-server",
                                "image": "quay.io/prometheus/prometheus:v2.36.2",
                                "imagePullPolicy": "IfNotPresent",
                                "securityContext": {},
                                "args": [
                                    "--storage.tsdb.retention.time=15d",
                                    "--config.file=/etc/config/prometheus.yml",
                                    "--storage.tsdb.path=/data",
                                    "--web.console.libraries=/etc/prometheus/console_libraries",
                                    "--web.console.templates=/etc/prometheus/consoles",
                                    "--web.enable-lifecycle",
                                ],
                                "ports": [{"containerPort": 9090}],
                                "readinessProbe": {
                                    "httpGet": {
                                        "path": "/-/ready",
                                        "port": 9090,
                                        "scheme": "HTTP",
                                    },
                                    "initialDelaySeconds": 30,
                                    "periodSeconds": 5,
                                    "timeoutSeconds": 4,
                                    "failureThreshold": 3,
                                    "successThreshold": 1,
                                },
                                "livenessProbe": {
                                    "httpGet": {
                                        "path": "/-/healthy",
                                        "port": 9090,
                                        "scheme": "HTTP",
                                    },
                                    "initialDelaySeconds": 30,
                                    "periodSeconds": 15,
                                    "timeoutSeconds": 10,
                                    "failureThreshold": 3,
                                    "successThreshold": 1,
                                },
                                "resources": {},
                                "volumeMounts": [
                                    {
                                        "name": "config-volume",
                                        "mountPath": "/etc/config",
                                    },
                                    {
                                        "name": "storage-volume",
                                        "mountPath": "/data",
                                        "subPath": "",
                                    },
                                ],
                            },
                        ],
                        "hostNetwork": False,
                        "dnsPolicy": "ClusterFirst",
                        "securityContext": {
                            "fsGroup": 65534,
                            "runAsGroup": 65534,
                            "runAsNonRoot": True,
                            "runAsUser": 65534,
                        },
                        "terminationGracePeriodSeconds": 300,
                        "volumes": [
                            {
                                "name": "config-volume",
                                "configMap": {"name": "jaseci-prometheus-server"},
                            },
                            {
                                "name": "storage-volume",
                                "persistentVolumeClaim": {
                                    "claimName": "jaseci-prometheus-server"
                                },
                            },
                        ],
                    },
                },
            },
        },
    ],
}



from redis import Redis

from jaseci.svc import CommonService
from .config import REDIS_CONFIG
from .manifest import REDIS_MANIFEST


#################################################
#                  REDIS HOOK                   #
#################################################


class RedisService(CommonService):

    ###################################################
    #                     BUILDER                     #
    ###################################################

    def run(self, hook=None):
        self.app = Redis(**self.config, decode_responses=True)
        self.app.ping()

    ###################################################
    #                     COMMONS                     #
    ###################################################

    def get(self, name):
        return self.app.get(name)

    def set(self, name, val):
        self.app.set(name, val)

    def exists(self, name):
        return self.app.exists(name)

    def delete(self, name):
        self.app.delete(name)

    def hget(self, name, key):
        return self.app.hget(name, key)

    def hset(self, name, key, val):
        self.app.hset(name, key, val)

    def hexists(self, name, key):
        return self.app.hexists(name, key)

    def hdel(self, name, key):
        self.app.hdel(name, key)

    def hkeys(self, name):
        return self.app.hkeys(name)

    ###################################################
    #                     CLEANER                     #
    ###################################################

    def clear(self):
        if self.is_running():
            self.app.flushdb()

    ###################################################
    #                     CONFIG                      #
    ###################################################

    def build_config(self, hook) -> dict:
        return hook.service_glob("REDIS_CONFIG", REDIS_CONFIG)

    def build_manifest(self, hook) -> dict:
        return hook.service_glob("REDIS_MANIFEST", REDIS_MANIFEST)


# ----------------------------------------------- #



from jaseci.actions.live_actions import jaseci_action
from jaseci.svc import MetaService


@jaseci_action()
def send(sender, recipients, subject, text, html):
    MetaService().get_service("mail").poke().send_custom_email(
        sender, recipients, subject, (text, html)
    )



import os


REDIS_CONFIG = {
    "enabled": True,
    "quiet": False,
    "host": os.getenv("REDIS_HOST", "localhost"),
    "port": os.getenv("REDIS_PORT", "6379"),
    "db": os.getenv("REDIS_DB", "1"),
}



from .state import ServiceState
from .common import CommonService, ProxyService, JsOrc, MetaProperties, Kube
from .redis import RedisService
from .task import TaskService
from .mail import MailService
from .prometheus import PrometheusService
from .elastic import ElasticService
from .meta import MetaService

__all__ = [
    "ServiceState",
    "JsOrc",
    "MetaProperties",
    "CommonService",
    "ProxyService",
    "RedisService",
    "TaskService",
    "MailService",
    "PrometheusService",
    "ElasticService",
    "MetaService",
    "Kube",
]



REDIS_MANIFEST = {
    "Service": [
        {
            "apiVersion": "v1",
            "kind": "Service",
            "metadata": {"name": "jaseci-redis"},
            "spec": {
                "selector": {"pod": "jaseci-redis"},
                "ports": [{"protocol": "TCP", "port": 6379, "targetPort": 6379}],
            },
        }
    ],
    "Deployment": [
        {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": {"name": "jaseci-redis"},
            "spec": {
                "replicas": 1,
                "selector": {"matchLabels": {"pod": "jaseci-redis"}},
                "template": {
                    "metadata": {"labels": {"pod": "jaseci-redis"}},
                    "spec": {
                        "containers": [
                            {
                                "name": "jaseci-redis-master",
                                "image": "redis",
                                "imagePullPolicy": "IfNotPresent",
                                "command": ["redis-server", "/redis-master/redis.conf"],
                                "resources": {"limits": {"cpu": "0.2"}},
                                "ports": [{"containerPort": 6379}],
                                "volumeMounts": [
                                    {"mountPath": "/redis-master-data", "name": "data"},
                                    {"mountPath": "/redis-master", "name": "config"},
                                ],
                            }
                        ],
                        "volumes": [
                            {"name": "data", "emptyDir": {}},
                            {
                                "name": "config",
                                "configMap": {
                                    "name": "jaseci-redis-config",
                                    "items": [
                                        {"key": "redis-config", "path": "redis.conf"}
                                    ],
                                },
                            },
                        ],
                    },
                },
            },
        }
    ],
    "ConfigMap": [
        {
            "apiVersion": "v1",
            "kind": "ConfigMap",
            "metadata": {"name": "jaseci-redis-config"},
            "data": {
                "redis-config": "maxmemory 1000mb\nmaxmemory-policy allkeys-lru\n"
            },
        }
    ],
}



"""
Prometheus APIs
"""
from jaseci.api.interface import Interface


class PrometheusApi:
    """
    Prometheus APIs
    """

    @Interface.admin_api()
    def prometheus_metrics_list(self):
        """
        Return list of availabel metrics
        """
        hook = self._h
        if hook.meta.run_svcs:
            return hook.promon.all_metrics()
        else:
            return {"success": False, "message": "No runnning Prometheus service."}

    @Interface.admin_api()
    def prometheus_pod_list(self, namespace: str = "", exclude_prom: bool = False):
        """
        Return list of pods. If exclude_prom,
        """
        hook = self._h
        if hook.meta.run_svcs:
            return hook.promon.pods(namespace=namespace, exclude_prom=exclude_prom)
        else:
            return {"success": False, "message": "No runnning Prometheus service."}

    @Interface.admin_api()
    def prometheus_pod_info(
        self,
        namespace: str = "",
        exclude_prom: bool = False,
        timestamp: int = 0,
        duration: int = 0,
    ):
        """
        Return pods info and metrics
        """
        hook = self._h
        if hook.meta.run_svcs:
            return hook.promon.info(
                namespace=namespace,
                exclude_prom=exclude_prom,
                timestamp=timestamp,
                duration=duration,
            )
        else:
            return {"success": False, "message": "No runnning Prometheus service."}
